"""
This module implements additional code for the server-side DeleteActor,
which is used to clean up the list of deleted nodes.
"""

from __future__ import annotations

import weakref
from collections import deque

from asyncactor import Actor, PingEvent, TagEvent
from asyncactor.backend import get_transport
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import anyio

TAGS = 4


class DeleteActor:
    _enabled = None

    def __init__(self, server):
        self._server = weakref.ref(server)
        self.deleted = deque()
        self.tags = []
        self.actor = None

        self.max_seen = 0
        self.n_tags = 0
        self.n_pings = 0
        self.n_nodes = 0

    @property
    def server(self):
        return self._server()

    async def tock_me(self):
        """
        Add the current tock to our buffer.

        This is updated whenever a new leader is selected.
        """
        self.tags.append(self.server.tock)
        self.tags = self.tags[-TAGS:]
        await self.actor.set_value((self.tags[0], self.tags[-1]))

    def add_deleted(self, nodes: NodeSet):  # noqa: F821
        """
        These nodes are deleted. Remember them for some time.
        """
        if self.n_nodes == 0:
            return
        self.deleted.append((self.server.tock, nodes))

    def purge_to(self, tock):
        """
        Sufficient time has passed since this tock was seen, while all
        Delete actor nodes were active. Finally flush the entries that have
        been deleted before it.
        """
        while self.deleted and self.deleted[0][0] < tock:
            d = self.deleted.popleft()
            self.server.purge_deleted(d[1])

    async def enable(self, n):
        """
        Enable this actor, as a group of N.
        """
        if self.actor is None:
            self._enabled = True
        else:
            await self.actor.enable(n)
        self.n_tags = 0
        self.n_pings = 0
        self.n_nodes = n

    async def disable(self, n: int = 0):
        """
        Disable this actor. It will still listen, and require N Delete
        actor members in order to flush its deletion entries.

        Completely disable deletion flushing by passing n=0.
        """
        if self.actor is None:
            self._enabled = False
        else:
            await self.actor.disable()
        self.n_tags = 0
        self.n_pings = 0
        self.n_nodes = n

    async def run(self, evt: anyio.abc.Event = None):
        """
        The task that monitors the Delete actor.
        """
        try:
            T = get_transport("moat_kv")
            async with Actor(
                T(self.server.backend, *self.server.cfg.server.root, "del"),
                name=self.server.node.name,
                cfg=self.server.cfg.server.delete,
                enabled=False,
            ) as actor:
                self.actor = actor
                if self._enabled is not None:
                    if self._enabled:
                        await actor.enable()
                    else:
                        await actor.disable()
                if evt is not None:
                    evt.set()
                async for evt in actor:
                    if isinstance(evt, PingEvent):
                        val = evt.value
                        if val is None:
                            self.n_pings = self.n_tags = 0
                            continue
                        if len(evt.msg.history) < self.n_nodes:
                            self.n_pings = self.n_tags = 0
                            continue
                        self.n_pings += 1
                        if self.n_pings > self.n_nodes:
                            mx, self.max_seen = (
                                self.max_seen,
                                max(self.max_seen, val[1]),
                            )
                            if val[0] > mx > 0:
                                await self.server.resync_deleted(evt.msg.history)
                                continue
                            self.purge_to(val[0])
                            self.max_seen = max(self.max_seen, val[1])

                    elif isinstance(evt, TagEvent):
                        if actor.history_size == self.n_nodes:
                            self.n_tags += 1
                            if self.n_tags > 2:
                                self.purge_to(self.tags[0])
                            await self.tock_me()
        finally:
            self.actor = None



#!/usr/bin/env python3
"""
Basic DistKV support

"""

from __future__ import annotations

import logging

import asyncclick as click
from moat.util import attrdict, combine_dict, load_subgroup, CFG, ensure_cfg

from moat.kv.auth import gen_auth
from moat.kv.client import client_scope

logger = logging.getLogger(__name__)


ensure_cfg("moat.kv")


class NullObj:
    """
    This helper defers raising an exception until one of its attributes is
    actually accessed.
    """

    def __init__(self, exc):
        self._exc = exc

    def __call__(self, *a, **kw):
        raise self._exc

    def __await__(self):
        raise self._exc

    def __getattr__(self, k):
        if k[0] == "_" and k not in ("_request", "_cfg"):
            return object.__getattribute__(self, k)
        raise self._exc


@load_subgroup(sub_pre="moat.kv.command", sub_post="cli", ext_pre="moat.kv", ext_post="_main.cli")
@click.option("-h", "--host", default=None, help=f"Host to use. Default: {CFG.kv.conn.host}")
@click.option(
    "-p",
    "--port",
    type=int,
    default=None,
    help=f"Port to use. Default: {CFG.kv.conn.port}",
)
@click.option(
    "-a",
    "--auth",
    type=str,
    default=None,
    help="Auth params. =file or 'type param=value…' Default: _anon",
)
@click.option("-m", "--metadata", is_flag=True, help="Include/print metadata.")
@click.pass_context
async def cli(ctx, host, port, auth, metadata):
    """The MoaT Key-Value subsystem.

    All commands (except 'server' and 'dump') connect to a MoaT-KV server.
    """
    obj = ctx.obj
    cfg = attrdict()
    if host is not None:
        cfg.host = host
    if port is not None:
        cfg.port = port

    if auth is not None:
        cfg.auth = gen_auth(auth)
        if obj.DEBUG:
            cfg.auth._DEBUG = True

    cfg = combine_dict(attrdict(kv=attrdict(conn=cfg)), obj.cfg, cls=attrdict)

    obj.meta = 3 if metadata else False

    try:
        if ctx.invoked_subcommand in {None, "server", "dump"}:
            obj.client = NullObj(RuntimeError("Not a client command"))
        else:
            obj.client = await client_scope(**cfg.kv)
    except OSError as exc:
        obj.client = NullObj(exc)
    else:
        logger.debug("Connected.")



"""
This module contains MoaT-KV's basic data model.

TODO: message chains should be refactored to arrays: much lower overhead.
"""

from __future__ import annotations

import weakref
from collections import defaultdict
from logging import getLogger
from typing import Any

from moat.util import NotGiven, Path, attrdict, create_queue
from range_set import RangeSet

from .exceptions import ACLError

logger = getLogger(__name__)


class NodeDataSkipped(Exception):
    def __init__(self, node):
        super().__init__()
        self.node = node

    def __repr__(self):
        return f"<{self.__class__.__name__}:{self.node}>"


ConvNull = None  # imported later, if/when needed
NullACL = None  # imported later, if/when needed


class Node:
    """Represents one MoaT-KV participant."""

    name: str = None
    tick: int = None
    _present: RangeSet = None  # I have these as valid data. Superset of ``._deleted``.
    _deleted: RangeSet = None  # I have these as no-longer-valid data
    _reported: RangeSet = None  # somebody else reported these missing data for this node
    _superseded: RangeSet = None  # I know these once existed, but no more.
    entries: dict = None
    tock: int = 0  # tock when node was last observed

    def __new__(cls, name, tick=None, cache=None, create=True):
        try:
            self = cache[name]
        except KeyError:
            if not create:
                raise
            self = object.__new__(cls)
            self.name = name
            self.tick = tick
            self._present = RangeSet()
            self._deleted = RangeSet()
            self._reported = RangeSet()
            self._superseded = RangeSet()
            self.entries = {}
            cache[name] = self
        else:
            if tick is not None and (self.tick is None or self.tick < tick):
                self.tick = tick
        return self

    # pylint: disable=unused-argument
    def __init__(self, name, tick=None, cache=None, create=True):
        return

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, Node):
            other = other.name
        return self.name == other

    def __getitem__(self, item):
        return self.entries[item]

    def get(self, item, default=None):
        return self.entries.get(item, default)

    def enumerate(self, n: int = 0, current: bool = False):
        """
        Return a list of valid keys for that node.

        Used to find data from no-longer-used nodes so they can be deleted.
        """
        for k, v in self.entries.items():
            if current and v.chain is not None and v.chain.node is not self:
                continue
            yield k
            if n:
                n -= 1
                if not n:
                    return

    def __contains__(self, item):
        return item in self.entries

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name} @{self.tick}>"

    def seen(self, tick, entry=None, local=False):
        """An event with this tick was in the entry's chain.

        Args:
          ``tick``: The event affecting the given entry.
          ``entry``: The entry affected by this event.
          ``local``: The message was not broadcast, thus do not assume that
                     other nodes saw this.

        """
        if not local:
            self._reported.discard(tick)
        if entry is not None:
            self._present.add(tick)
            self.entries[tick] = entry

            # this might happen when loading old state.
            if tick in self._superseded:
                self._superseded.discard(tick)
                logger.info("%s was marked as superseded", entry)

    def is_deleted(self, tick):
        """
        Check whether this tick has been marked as deleted.
        """
        return tick in self._deleted

    def mark_deleted(self, tick):
        """
        The data for this tick will be deleted.

        Args:
          tick: The event that caused the deletion.

        Returns: the entry, if still present
        """
        self._deleted.add(tick)
        return self.entries.get(tick, None)

    def clear_deleted(self, tick):
        """
        The data for this tick are definitely gone (deleted).
        """
        self._present.discard(tick)
        self._deleted.discard(tick)
        self._reported.discard(tick)
        self._superseded.add(tick)

        e = self.entries.pop(tick, None)
        if e is not None:
            e.purge_deleted()

    def purge_deleted(self, r: RangeSet):
        """
        All entries in this rangeset are deleted.

        This is a shortcut for calling :meth:`clear_deleted` on each item.
        """
        self._present -= r
        self._deleted -= r
        self._reported -= r
        self._superseded += r

        # Mark that these as really gone.
        for a, b in r:
            for t in range(a, b):
                e = self.entries.pop(t, None)
                if e is not None:
                    e.purge_deleted()

    def supersede(self, tick):
        """
        The event with this tick is no longer in the referred entry's chain.
        This happens when an entry is updated.

        Args:
          ``tick``: The event that once affected the given entry.
        """
        self._present.discard(tick)
        self._superseded.add(tick)
        self.entries.pop(tick, None)

    def report_superseded(self, r: RangeSet, local=False):
        """
        Some node said that these entries may have been superseded.

        Args:
          ``range``: The RangeSet thus marked.
          ``local``: The message was not broadcast, thus do not assume that
                     other nodes saw this.
        """
        coll = r & self._present
        r -= coll
        self._superseded += r
        if not local:
            self._reported -= r

        # Are any of these present?
        for a, b in coll:
            for t in range(a, b):
                e = self.entries[t]
                if e.chain.node is self and e.chain.tick == t:
                    logger.info("%s present but marked as superseded", e)

    def report_missing(self, r: RangeSet):
        """
        Some node doesn't know about these ticks.

        We may need to broadcast either their content,
        or the fact that these ticks have been superseded.
        """
        self._reported += r

    def report_deleted(self, r: RangeSet, server):
        """
        This range has been reported as deleted.

        Args:
          range (RangeSet): the range that's gone.
          add (dict): store additional vanished items. Nodename -> RangeSet
        """
        # Ignore those which we know have been deleted
        n = r - self._deleted

        for a, b in n:
            for t in range(a, b):
                entry = self.mark_deleted(t)
                if entry is None:
                    continue
                chain = entry.mark_deleted(server)
                if chain is None:
                    continue
                for node, tick in chain:
                    server.mark_deleted(node, tick)

        # Mark as deleted. The system will flush them later.
        self._deleted += r
        self._reported -= r
        self._superseded += r

        r &= self._present
        for a, b in r:
            for t in range(a, b):
                e = self.entries.get(t, None)
                if e is not None:
                    if e.data is not NotGiven:
                        logger.info("%s present but marked as deleted", e)
                    e.purge_deleted()

    @property
    def local_present(self):
        """Values I know about"""
        return self._present

    @property
    def local_superseded(self):
        """Values I knew about"""
        return self._superseded

    @property
    def local_deleted(self):
        """Values I know to have vanished"""
        return self._deleted

    @property
    def local_missing(self):
        """Values I have not seen, the inverse of :meth:`local_present`
        plus :meth:`local_superseded`"""
        assert self.tick
        r = RangeSet(((1, self.tick + 1),))
        r -= self._present
        r -= self._superseded
        return r

    @property
    def remote_missing(self):
        """Values from this node which somebody else has not seen"""
        return self._reported

    def kill_this_node(self, cache=None):
        """
        Remove this node from the system.
        No chain's first link may point to this node.
        """
        for e in self.entries.values():
            if e.chain.node is self:
                raise RuntimeError(f"Still main node at {e!r}")
            c = e.chain.filter(self)
            if c is None:
                raise RuntimeError(f"Empty chain after filter for {self.name} at {e!r}")
            e.chain = c

        if cache is not None:
            cache.pop(self.name, None)


class NodeSet(defaultdict):
    """
    Represents a dict (nodename > RangeSet).
    """

    def __init__(self, encoded=None, cache=None):
        super().__init__(RangeSet)
        if encoded is not None:
            assert cache is not None
            for n, v in encoded.items():
                n = Node(n, cache=cache)
                r = RangeSet()
                r.__setstate__(v)
                self[n] = r

    def __repr__(self):
        return f"{type(self).__name__}({dict.__repr__(self)})"

    def __bool__(self):
        return any(v for v in self.values())

    def serialize(self):
        d = dict()
        for k, v in self.items():
            assert not hasattr(k, "name")
            d[k] = v.__getstate__()
        return d

    @classmethod
    def deserialize(cls, state):
        self = cls()
        for k, v in state.items():
            r = RangeSet()
            r.__setstate__(v)
            self[k] = v
        return self

    def copy(self):
        res = type(self)()
        for k, v in self.items():
            if v:
                res[k] = v.copy()
        return res

    def add(self, node, tick):
        assert not hasattr(node, "name")
        self[node].add(tick)

    def __isub__(self, other):
        for k, v in other.items():
            r = self.get(k, None)
            if r is None:
                continue
            r -= v


class NodeEvent:
    """Represents any event originating at a node.

    Args:
      ``node``: The node thus affected
      ``tick``: Counter, timestamp, whatever
      ``prev``: The previous event, if any

    """

    def __init__(self, node: Node, tick: int | None = None, prev: NodeEvent = None):
        self.node = node
        if tick is None:
            tick = node.tick
        self.tick = tick
        if tick is not None and tick > 0:
            node.seen(tick)
        self.prev = None
        if prev is not None:
            self.prev = prev

    def __len__(self):
        """Length of this chain"""
        if self.prev is None:
            return 1
        return 1 + len(self.prev)

    def _repr_chain(self):
        tk = self.tick or "-"
        pr = f"{self.node.name}:{tk}"
        if self.prev is not None:
            pr += " " + self.prev._repr_chain()
        return pr

    def __repr__(self):
        pr = self._repr_chain()
        tk = ("-" if self.tick is None else self.tick,)
        return f"<{self.__class__.__name__} {pr} {tk}>"

    def __iter__(self):
        c = self
        while c is not None:
            yield c.node, c.tick
            c = c.prev

    def __eq__(self, other):
        if other is None:
            return False
        return self.node == other.node and self.tick == other.tick

    def equals(self, other):
        """Check whether these chains are equal. Used for ping comparisons.

        The last two items may be missing from either chain.
        """
        if other is None:
            return self.prev is None or len(self.prev) <= 1
        if self != other:
            return False
        if self.prev is None:
            return other.prev is None or len(other.prev) <= 2
        elif other.prev is None:
            return self.prev is None or len(self.prev) <= 2
        else:
            return self.prev.equals(other.prev)

    def __lt__(self, other):
        """Check whether this node precedes ``other``, i.e. "other" is an
        event that is based on this.
        """
        if other is None:
            return False
        if self == other:
            return False
        while self.node != other.node:
            other = other.prev
            if other is None:
                return False
        return self.tick <= other.tick

    def __gt__(self, other):
        """Check whether this node succedes ``other``, i.e. this event is
        based on it.
        """
        if other is None:
            return True
        if self == other:
            return False
        while self.node != other.node:
            self = self.prev  # pylint: disable=self-cls-assignment
            if self is None:
                return False
        return self.tick >= other.tick

    def __lte__(self, other):
        return self.__eq__(other) or self.__lt__(other)

    def __gte__(self, other):
        return self.__eq__(other) or self.__gt__(other)

    def __contains__(self, node):
        return self.find(node) is not None

    def find(self, node):
        """Return the position of a node in this chain.
        Zero if the first entry matches.

        Returns ``None`` if not present.
        """
        res = 0
        while self is not None:
            if self.node == node:
                return res
            res += 1
            self = self.prev  # pylint: disable=self-cls-assignment
        return None

    def filter(self, node, server=None):
        """Return an event chain without the given node.

        If the node is not in the chain, the result is *not* a copy.
        """
        if self.node == node:
            if server is not None:
                server.drop_old_event(self)
            return self.prev
            # Invariant: a node can only be in the chain once
            # Thus we can stop filtering after we encounter it.
        if self.prev is None:
            return self
        prev = self.prev.filter(node, server=server)
        if prev is self.prev:
            # No change, so return unmodified
            return self
        return NodeEvent(node=self.node, tick=self.tick, prev=prev)

    def serialize(self, nchain=-1) -> dict:
        if not nchain:
            raise RuntimeError("A chopped-off NodeEvent must not be sent")
        res = attrdict(node=self.node.name)
        if self.tick is not None:
            res.tick = self.tick
        if self.prev is None:
            res.prev = None
        elif nchain != 1:
            res.prev = self.prev.serialize(nchain - 1)
        return res

    @classmethod
    def deserialize(cls, msg, cache):
        if msg is None:
            return None
        msg = msg.get("chain", msg)
        tick = msg.get("tick", None)
        if "node" not in msg:
            assert "prev" not in msg
            assert tick is None
            return None
        self = cls(node=Node(msg["node"], tick=tick, cache=cache), tick=tick)
        if "prev" in msg:
            self.prev = cls.deserialize(msg["prev"], cache=cache)
        return self

    def attach(self, prev: NodeEvent = None, server=None):
        """Copy this node, if necessary, and attach a filtered `prev` chain to it"""
        if prev is not None:
            prev = prev.filter(self.node, server=server)
        if self.prev is not None or prev is not None:
            self = NodeEvent(  # pylint: disable=self-cls-assignment
                node=self.node,
                tick=self.tick,
                prev=prev,
            )
        return self


class UpdateEvent:
    """Represents an event which updates something."""

    def __init__(self, event: NodeEvent, entry: Entry, new_value, old_value=NotGiven, tock=None):
        self.event = event
        self.entry = entry
        self.new_value = new_value
        if old_value is not NotGiven:
            self.old_value = old_value
        if new_value is NotGiven:
            event.node.mark_deleted(event.tick)
        self.tock = tock

    def __repr__(self):
        if self.entry.chain == self.event:
            res = ""
        else:
            res = repr(self.event) + ": "
        return "<{}:{}{}: {}→{}>".format(
            self.__class__.__name__,
            res,
            repr(self.entry),
            "-"
            if not hasattr(self, "old_value")
            else ""
            if self.new_value == self.entry.data
            else repr(self.old_value),
            repr(self.new_value),
        )

    def serialize(self, chop_path=0, nchain=-1, with_old=False, conv=None):
        if conv is None:
            # pylint: disable=redefined-outer-name
            global ConvNull
            if ConvNull is None:
                from .types import ConvNull
            conv = ConvNull
        res = self.event.serialize(nchain=nchain)
        res.path = self.entry.path[chop_path:]
        if with_old:
            if self.old_value is not NotGiven:
                res.old_value = conv.enc_value(self.old_value, entry=self.entry)
            if self.new_value is not NotGiven:
                res.new_value = conv.enc_value(self.new_value, entry=self.entry)
        elif self.new_value is not NotGiven:
            res.value = conv.enc_value(self.new_value, entry=self.entry)
        res.tock = self.entry.tock
        return res

    @classmethod
    def deserialize(cls, root, msg, cache, nulls_ok=False, conv=None):
        if conv is None:
            # pylint: disable=redefined-outer-name
            global ConvNull
            if ConvNull is None:
                from .types import ConvNull
            conv = ConvNull
        entry = root.follow(msg.path, create=True, nulls_ok=nulls_ok)
        event = NodeEvent.deserialize(msg, cache=cache)
        old_value = NotGiven
        if "value" in msg:
            value = conv.dec_value(msg.value, entry=entry)
        elif "new_value" in msg:
            value = conv.dec_value(msg.new_value, entry=entry)
            old_value = conv.dec_value(msg.old_value, entry=entry)
        else:
            value = NotGiven

        return cls(event, entry, value, old_value=old_value, tock=msg.tock)


class Entry:
    """This class represents one key/value pair"""

    _parent: Entry = None
    name: str = None
    _path: list[str] = None
    _root: Entry = None
    chain: NodeEvent = None
    SUBTYPE = None
    SUBTYPES = {}
    _data: Any = NotGiven

    monitors = None

    def __init__(self, name: str, parent: Entry, tock=None):
        self.name = name
        self._sub = {}
        self.monitors = set()
        self.tock = tock

        if parent is not None:
            parent._add_subnode(self)
            self._parent = weakref.ref(parent)

    def _add_subnode(self, child: Entry):
        self._sub[child.name] = child

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, Entry):
            other = other.name
        return self.name == other

    def chain_links(self):
        c = self.chain
        if c is not None:
            yield from iter(c)

    def keys(self):
        return self._sub.keys()

    def values(self):
        return self._sub.values()

    def items(self):
        return self._sub.items()

    def __len__(self):
        return len(self._sub)

    def __bool__(self):
        return self._data is not None or len(self._sub) > 0

    def __contains__(self, key):
        return key in self._sub

    @property
    def path(self):
        if self._path is None:
            parent = self.parent
            if parent is None:
                self._path = Path()
            else:
                self._path = parent.path + [self.name]
        return self._path

    def follow_acl(self, path, *, create=True, nulls_ok=False, acl=None, acl_key=None):
        """Follow this path.

        If ``create`` is True (default), unknown nodes are silently created.
        Otherwise they cause a `KeyError`. If ``None``, assume
        ``create=True`` but only check the ACLs.

        If ``nulls_ok`` is False (default), `None` is not allowed as a path
        element. If 2, it is allowed anywhere; if True, only as the first
        element.

        If ``acl`` is not ``None``, then ``acl_key`` is the ACL letter to
        check for. ``acl`` must be an :class:`~moat.kv.types.ACLFinder`
        created from the root of the ACL in question.

        The ACL key 'W' is special: it checks 'c' if the node is new, else
        'w'.

        Returns a (node, acl) tuple.
        """

        # KEEP IN SYNC with `follow`, below!
        if acl is None:
            global NullACL
            if NullACL is None:
                from .types import NullACL  # pylint: disable=redefined-outer-name
            acl = NullACL

        first = True
        for name in path:
            if name is None and not nulls_ok:
                raise ValueError("Null path element")
            if nulls_ok == 1:  # root only
                nulls_ok = False
            child = self._sub.get(name, None) if self is not None else None
            if child is None:
                if create is False:
                    raise KeyError(path)
                acl.check("n")
                if create is not None:
                    child = self.SUBTYPES.get(name, self.SUBTYPE)
                    if child is None:
                        raise ValueError(f"Cannot add {name} to {self}")
                    child = child(name, self, tock=self.tock)
            else:
                acl.check("x")
            try:
                acl = acl.step(name, new=first)
            except KeyError:
                raise ACLError(acl.result, name) from None
            first = False
            self = child  # pylint: disable=self-cls-assignment

        # If the caller doesn't know if the node exists, help them out.
        if acl_key == "W":
            acl_key = "w" if self is not None and self._data is not NotGiven else "c"
        acl.check(acl_key)
        return (self, acl)

    def follow(self, path, *, create=True, nulls_ok=False):
        """
        As :meth:`follow_acl`, but isn't interested in ACLs and only returns the node.
        """
        # KEEP IN SYNC with `follow_acl`, above!

        for name in path:
            if name is None and not nulls_ok:
                raise ValueError("Null path element")
            if nulls_ok == 1:  # root only
                nulls_ok = False
            child = self._sub.get(name, None) if self is not None else None
            if child is None:
                if create is False:
                    raise KeyError(path)
                if create is not None:
                    child = self.SUBTYPES.get(name, self.SUBTYPE)
                    if child is None:
                        raise ValueError(f"Cannot add {name} to {self}")
                    child = child(name, self, tock=self.tock)
            self = child  # pylint: disable=self-cls-assignment
        return self

    def __getitem__(self, name):
        return self._sub[name]

    @property
    def root(self):
        root = self._root
        if root is not None:
            root = root()
            if root is None:
                raise RuntimeError("Root node is gone")
            return root

        parent = self.parent
        if parent is None:
            return self
        root = parent.root
        self._root = weakref.ref(root)
        return root

    async def set(self, value):
        self._data = value

    @property
    def parent(self):
        parent = self._parent
        if parent is None:
            return None
        parent = parent()
        if parent is None:
            raise RuntimeError("Parent node is gone")
        return parent

    def __repr__(self):
        try:
            res = f"<{self.__class__.__name__}:{self.path}"
            if self.chain is not None:
                res += f"@{self.chain!r}"
            if self.data is not None:
                res += f" ={self.data!r}"
            res += ">"
        except Exception as exc:
            res = f"<{self.__class__.__name__}:{exc!s}"
        return res

    @property
    def data(self):
        return self._data

    def mark_deleted(self, server):
        """
        This entry has been deleted.

        Returns:
          the entry's chain.
        """
        c = self.chain
        for node, tick in self.chain_links():
            e = node.mark_deleted(tick)
            assert e is None or e is self
            server.mark_deleted(node, tick)
        self._data = NotGiven
        return c

    def purge_deleted(self):
        """
        Call :meth:`Node.clear_deleted` on each link in this entry's chain.
        """
        c, self.chain = self.chain, None
        if c is None:
            return
        for node, tick in c:
            node.clear_deleted(tick)
        self._chop()

    def _chop(self):
        """
        Remove a deleted entry (and possibly its parent).
        """
        logger.debug("CHOP %r", self)
        this, p = self, self._parent
        while p is not None:
            p = p()
            if p is None:
                return
            p._sub.pop(this.name, None)
            if p._sub:
                return
            this, p = p, p._parent

    async def set_data(self, event: NodeEvent, data: Any, server=None, tock=None):
        """This entry is updated by that event.

        Args:
          event: The :class:`NodeEvent` to base the update on.
          data (Any): whatever the node should contains. Use :any:`moat.kv.util.NotGiven`
            to delete.

        Returns:
          The :class:`UpdateEvent` that has been generated and applied.
        """
        event = event.attach(self.chain, server=server)
        evt = UpdateEvent(event, self, data, self._data, tock=tock)
        await self.apply(evt, server=server)
        return evt

    async def apply(self, evt: UpdateEvent, server=None, root=None, loading=False):
        """Apply this :cls`UpdateEvent` to me.

        Also, forward to watchers.
        """
        chk = None
        if root is not None and None in root:
            chk = root[None].get("match", None)

        if evt.event is None:
            raise RuntimeError("huh?")

        if evt.event == self.chain:
            if False and self._data != evt.new_value:
                # codec diffs due to wrong accurracy
                logger.error(
                    "Diff %r: has\n%r\nbut should have\n%r\n",
                    evt.event,
                    self._data,
                    evt.new_value,
                )
            return

        if hasattr(evt, "new_value"):
            evt_val = evt.new_value
        else:
            evt_val = evt.value

        if self.chain > evt.event:  # already superseded.
            logger.info("*** superseded ***")
            logger.info("Node: %s", self.path)
            logger.info("Current: %s :%s: %r", self.chain, self.tock, self._data)
            logger.info("New: %s :%s: %r", evt.event, evt.tock, evt_val)
            return

        if self._data is not NotGiven and not (self.chain < evt.event):
            if not loading:
                logger.warning("*** inconsistency ***")
                logger.warning("Node: %s", self.path)
                logger.warning("Current: %s :%s: %r", self.chain, self.tock, self._data)
                logger.warning("New: %s :%s: %r", evt.event, evt.tock, evt_val)
            if evt.tock < self.tock:
                if not loading:
                    logger.warning("New value ignored")
                # also mark the new event's chain as superseded
                server.drop_old_event(self.chain, evt.event)
                return
            if not loading:
                logger.warning("New value used")

        if chk is not None and evt_val is not NotGiven:
            chk.check_value(evt_val, self)
        if not hasattr(evt, "old_value"):
            evt.old_value = self._data
        await self.set(evt_val)
        self.tock = evt.tock

        server.drop_old_event(evt.event, self.chain)
        self.chain = evt.event

        if evt_val is NotGiven:
            self.mark_deleted(server)

        for n, t in self.chain_links():
            n.seen(t, self)
        await self.updated(evt)

    async def walk(self, proc, acl=None, max_depth=-1, min_depth=0, _depth=0, full=False):
        """
        Call coroutine ``proc`` on this node and all its children).

        If `acl` (must be an ACLStepper) is given, `proc` is called with
        the acl as second argument.

        If `proc` raises `StopAsyncIteration`, chop this subtree.
        """
        if min_depth <= _depth:
            try:
                if acl is not None:
                    await proc(self, acl)
                else:
                    await proc(self)
            except StopAsyncIteration:
                return
        if max_depth == _depth:
            return
        _depth += 1
        for k, v in list(self._sub.items()):
            if k is None and not full:
                continue
            a = acl.step(k) if acl is not None else None
            await v.walk(proc, acl=a, max_depth=max_depth, min_depth=min_depth, _depth=_depth)

    def serialize(self, chop_path=0, nchain=2, conv=None):
        """Serialize this entry for msgpack.

        Args:
          ``chop_path``: If <0, do not return the entry's path.
                         Otherwise, do, but remove the first N entries.
          ``nchain``: how many change events to include.
        """
        if conv is None:
            global ConvNull
            if ConvNull is None:
                from .types import ConvNull  # pylint: disable=redefined-outer-name
            conv = ConvNull
        res = attrdict()
        if self._data is not NotGiven:
            res.value = conv.enc_value(self._data, entry=self)
        if self.chain is not None and nchain != 0:
            res.chain = self.chain.serialize(nchain=nchain)
        res.tock = self.tock
        if chop_path >= 0:
            path = self.path
            if chop_path > 0:
                path = path[chop_path:]
            res.path = path
        return res

    async def updated(self, event: UpdateEvent):
        """Send an event to this node (and all its parents)'s watchers."""
        node = self
        while True:
            bad = set()
            for q in list(node.monitors):
                if q._moat.kv__free is None or q._moat.kv__free > 1:
                    if q._moat.kv__free is not None:
                        q._moat.kv__free -= 1
                    await q.put(event)
                else:
                    bad.add(q)
            for q in bad:
                try:
                    if q._moat.kv__free > 0:
                        await q.put(None)
                    node.monitors.remove(q)
                except KeyError:
                    pass
                else:
                    pass
                    # await q.aclose()
            node = node.parent
            if node is None:
                break

    _counter = 0

    @property
    def counter(self):
        self._counter += 1
        return self._counter


Entry.SUBTYPE = Entry


class Watcher:
    """
    This helper class is used as an async context manager plus async
    iterator. It reports all updates to an entry (or its children).

    If a watcher terminates, sending to its channel has blocked.
    The receiver needs to take appropriate re-syncing action.
    """

    root: Entry = None
    q = None
    q_len = 100

    def __init__(self, root: Entry, full: bool = False, q_len: int | None = None):
        self.root = root
        self.full = full
        if q_len is not None:
            self.q_len = q_len

    async def __aenter__(self):
        if self.q is not None:
            raise RuntimeError("You cannot enter this context more than once")
        self.q = create_queue(self.q_len)
        self.q._moat.kv__free = self.q_len or None  # pylint:disable=no-member
        self.root.monitors.add(self.q)
        return self

    async def __aexit__(self, *tb):
        self.root.monitors.remove(self.q)
        self.q = None

    def __aiter__(self):
        if self.q is None:
            raise RuntimeError("You need to enclose this with 'async with'")
        return self

    async def __anext__(self):
        if self.q is None:
            raise RuntimeError("Aborted. Queue filled?")
        while True:
            res = await self.q.get()
            if self.q._moat.kv__free is not None:  # pylint:disable=no-member
                self.q._moat.kv__free += 1  # pylint:disable=no-member
            if res is None:
                raise RuntimeError("Aborted. Queue filled?")
            if len(res.entry.path) and res.entry.path[0] is None and not self.full:
                continue
            return res



"""
Object interface to moat.kv data

"""

from __future__ import annotations

import heapq
import weakref
from collections.abc import Mapping
from functools import partial

import anyio
from asyncscope import scope

try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

from moat.util import (
    NoLock,
    NotGiven,
    Path,
    PathLongener,
    combine_dict,
    yload,
    ensure_cfg,
    CFG,
)
import contextlib

__all__ = ["ClientEntry", "AttrClientEntry", "ClientRoot"]


class NamedRoot:
    """
    This is a mix-on class for the root of a subhierarchy that caches named
    sub-entries.

    Named children should call `_add_name` on this entry.
    """

    def __init__(self, *a, **k):
        self.__named = {}
        super().__init__(*a, **k)

    def by_name(self, name):
        if name is None:
            return None
        if not isinstance(name, str):
            raise ValueError("No string: " + repr(name))
        return self.__named.get(name)

    def _add_name(self, obj):
        n = obj.name
        if n is None:
            return

        self.__named[n] = obj
        obj.reg_del(self, "_del__name", obj, n)

    def _del__name(self, obj, n):
        old = self.__named.pop(n)
        if old is None or old is obj:
            return
        # Oops, that has been superseded. Put it back.
        self.__named[n] = old


class ClientEntry:
    """A helper class that represents a node on the server, as returned by
    :meth:`Client.mirror`.
    """

    value = NotGiven
    chain = None

    def __init__(self, parent, name=None):
        self._init()
        self._path = parent._path + (name,)
        self._name = name
        self._parent = weakref.ref(parent)
        self._root = weakref.ref(parent.root)
        self.client = parent.client

    def _init(self):
        self._lock = anyio.Lock()  # for saving etc.
        self.chain = None
        self.value = NotGiven
        self._children = dict()

    @classmethod
    def child_type(cls, name):  # pylint: disable=unused-argument
        """Given a node, return the type which the child with that name should have.
        The default is "same as this class".
        """
        return cls

    def value_or(self, default, typ=None):
        """
        Shortcut to coerce the value to some type
        """
        val = self.value
        if val is NotGiven:
            return default
        if typ is not None and not isinstance(val, typ):
            return default
        return val

    def val(self, *attr):
        """
        Shortcut to get an attribute value
        """
        return self.val_d(NotGiven, *attr)

    def val_d(self, default, *attr):
        """
        Shortcut to get an attribute value, or a default
        """
        val = self.value
        if val is NotGiven:
            if default is NotGiven:
                raise ValueError("no value set")
            return default
        for a in attr:
            try:
                val = val[a]
            except KeyError:
                if default is NotGiven:
                    raise
                return default
        return val

    def find_cfg(self, *k, default=NotGiven):
        """
        Convenience method to get a config value

        It is retrieved first from this node's value, then from the parent.
        """
        val = self.value_or({}, Mapping)
        try:
            for kk in k:
                try:
                    val = val[kk]
                except TypeError:
                    raise TypeError(self.value, k, val, kk)
            return val
        except KeyError:
            return self.parent.find_cfg(*k, default=default)

    @property
    def parent(self):
        return self._parent()

    @property
    def root(self):
        return self._root()

    @property
    def subpath(self):
        """Return the path to this entry, starting with its :class:`ClientRoot` base."""
        return self._path[len(self.root._path) :]

    @property
    def all_children(self):
        """Iterate all child nodes with data.

        You can send ``True`` to the iterator if you want to skip a subtree.
        """
        for k in self:
            if k.value is not NotGiven:
                res = yield k
                if res is True:
                    continue
            yield from k.all_children

    def allocate(self, name: str, exists: bool = False):
        """
        Create the child named "name". It is created (locally) if it doesn't exist.

        Arguments:
          name: The child node's name.
          exists: return the existing value? otherwise error

        If this returns ``None``, the subtree shall not be tracked.

        """
        c = self._children.get(name, None)
        if c is not None:
            if exists:
                return c
            raise RuntimeError("Duplicate child", self, name, c)
        c = self.child_type(name)
        if c is None:
            raise KeyError(name)
        self._children[name] = c = c(self, name)
        return c

    def follow(self, path, *, create=None, empty_ok=False):
        """Look up a sub-entry.

        Arguments:
          path (Path): the path elements to follow.
          create (bool): Create the entries. Default ``False``.
            Otherwise return ``None`` if not found.

        The path may not be empty. It also must not be a string,
        because that indicates that you called ``.follow(*path)`` instead of
        ``.follow(path)``.
        """
        if isinstance(path, str):
            raise RuntimeError("You seem to have used '*path' instead of 'path'.")
        if not empty_ok and not len(path):
            raise RuntimeError("Empty path")

        node = self
        for n, elem in enumerate(path, start=1):
            next_node = node.get(elem)
            if next_node is None:
                if create is False:
                    return None
                next_node = node.allocate(elem)
            elif create and n == len(path) and next_node.value is not NotGiven:
                raise RuntimeError("Duplicate child", self, path, n)

            node = next_node

        return node

    def __getitem__(self, name):
        return self._children[name]

    def by_name(self, name):
        """
        Lookup by a human-readable name?
        """
        return self._children[name]

    def __delitem__(self, name):
        del self._children[name]

    def get(self, name):
        return self._children.get(name, None)

    def __iter__(self):
        """Iterating an entry returns its children."""
        return iter(list(self._children.values()))

    def __bool__(self):
        return self.value is not NotGiven or bool(self._children)

    def __len__(self):
        return len(self._children)

    def __contains__(self, k):
        if isinstance(k, type(self)):
            k = k._name
        return k in self._children

    async def update(self, value, _locked=False, wait=False):
        """Update (or simply set) this node's value.

        This is a coroutine.
        """
        async with NoLock if _locked else self._lock:
            r = await self.root.client.set(
                self._path,
                chain=self.chain,
                value=value,
                nchain=3,
                idem=True,
            )
            if wait:
                await self.root.wait_chain(r.chain)
            self.value = value
            self.chain = r.chain
            return r

    async def delete(self, _locked=False, nchain=0, chain=True, wait=False, recursive=False):
        """Delete this node's value.

        This is a coroutine.
        """
        async with NoLock if _locked else self._lock:
            r = await self.root.client.delete(
                self._path,
                nchain=nchain,
                recursive=recursive,
                **({"chain": self.chain} if chain else {}),
            )
            if wait:
                await self.root.wait_chain(r.chain)
            self.chain = None
            return r

    async def set_value(self, value):
        """Callback to set the value when data has arrived.

        This method is strictly for overriding.
        Don't call me, I'll call you.

        This is a coroutine, for ease of integration.
        """
        self.value = value

    async def seen_value(self):
        """The current value was seen (again).

        Useful for syncing.

        The default action is to do nothing.
        """
        pass

    def mark_inconsistent(self, r):
        """There has been an inconsistent update.

        This call will immediately be followed by a call to
        :meth:`set_value`, thus it is not async.

        The default action is to do nothing.
        """
        pass


def _node_gt(self, other):
    if other is None and self is not None:
        return True
    if self is None or self == other:
        return False
    while self["node"] != other["node"]:
        self = self["prev"]
        if self is None:
            return False
    return self["tick"] >= other["tick"]


class AttrClientEntry(ClientEntry):
    """A ClientEntry which expects a dict as value and sets (some of) the clients'
    attributes appropriately.

    Set the classvar ``ATTRS`` to a list of the attrs you want saved. Note
    that these are not inherited: when you subclass, copy and extend the
    ``ATTRS`` of your superclass.

    If the entry is deleted (value set to ``None``, the attributes listed in
    ``ATTRS`` will be deleted too, or revert to the class values.
    """

    ATTRS = ()

    async def update(self, value, **kw):  # pylint: disable=arguments-differ
        raise RuntimeError("Nope. Set attributes and call '.save()'.")

    async def set_value(self, value):
        """Callback to set the value when data has arrived.

        This method sets the actual attributes.

        This method is strictly for overriding.
        Don't call me, I'll call you.
        """
        await super().set_value(value)
        for k in self.ATTRS:
            if value is not NotGiven and k in value:
                setattr(self, k, value[k])
            else:
                with contextlib.suppress(AttributeError):
                    delattr(self, k)

    def get_value(self, skip_none=False, skip_empty=False):
        """
        Extract value from attrs
        """
        res = {}
        for attr in type(self).ATTRS:
            try:
                v = getattr(self, attr)
            except AttributeError:
                pass
            else:
                if v is NotGiven:
                    continue
                if skip_none and v is None:
                    continue
                if skip_empty and v == ():
                    continue
                res[attr] = v
        return res

    async def save(self, wait=False):
        """
        Save myself to storage, by copying ATTRS to a new value.
        """
        async with self._lock:
            r = await super().update(value=self.get_value(), _locked=True)
            if wait:
                await self.root.wait_chain(r.chain)
            return r


class MirrorRoot(ClientEntry):
    """
    This class represents the root of a subsystem's storage, used for
    object-agnostic data mirroring.

    Used internally.
    """

    _tg = None

    CFG = None  # You need to override this with a dict(prefix=('where','ever'))

    def __init__(self, client, path, *, need_wait=False, cfg=None, require_client=True):
        # pylint: disable=super-init-not-called
        self._init()
        self.client = client
        self._path = path
        self._need_wait = need_wait
        self._loaded = anyio.Event()
        self._require_client = require_client

        if cfg is None:
            cfg = {}
        self._cfg = cfg
        self._name = self.client.name

        if need_wait:
            self._waiters = dict()
            self._seen = dict()

    @classmethod
    async def as_handler(cls, client, cfg=None, key="prefix", subpath=(), name=None, **kw):
        """Return an (or "the") instance of this class.

        The handler is created if it doesn't exist.

        Instances are distinguished by a key (from config), which
        must contain their path, and an optional subpath.
        """
        d = []
        if cfg is not None:
            d.append(cfg)

        defcfg = client._cfg.get(cls.CFG)
        if not defcfg:
            # seems we didn't load the class' default config yet.
            import inspect
            from pathlib import Path as _Path

            md = inspect.getmodule(cls)
            ensure_cfg("moat.kv")
            defcfg = CFG.kv.get(cls.CFG)
            try:
                f = (_Path(md.__file__).parent / "_nconfig.yaml").open("r")
            except OSError:
                pass
            else:
                with f:
                    defcfg = yload(f, attr=True).get("kv", {}).get(cls.CFG)
        if cfg:
            if defcfg:
                cfg = combine_dict(cfg, defcfg)
        else:
            if not defcfg:
                raise RuntimeError("no config for " + repr(cls))
            cfg = defcfg

        if name is None:
            # if key != "prefix":
            #     subpath = Path(key) + subpath
            name = str(Path("_moat.kv", client.name, cls.CFG, *subpath))

        def make():
            return client.mirror(cfg[key] + subpath, root_type=cls, need_wait=True, cfg=cfg, **kw)

        return await client.unique_helper(name, factory=make)

    @classmethod
    def child_type(cls, name):
        """Given a node, return the type which the child with that name should have.
        The default is :class:`ClientEntry`.

        This may return ``None``. In that case the subtree with this name
        shall not be tracked further.
        """
        return ClientEntry

    @property
    def root(self):
        """Returns this instance."""
        return self

    def find_cfg(self, *k, default=NotGiven):
        """
        Convenience method to get a config value.

        It is retrieved first from this node's value, then from the configuration read via CFG.
        """
        val = self.value_or({}, Mapping)
        try:
            for kk in k:
                val = val[kk]
            return val
        except KeyError:
            try:
                val = self._cfg
                for kk in k:
                    val = val[kk]
                return val
            except KeyError:
                if default is NotGiven:
                    raise
                return default

    async def run_starting(self):
        """Hook for 'about to start reading'"""
        pass

    async def running(self):
        """Hook for 'done reading current state'"""
        self._loaded.set()

    @asynccontextmanager
    async def run(self):
        """A coroutine that fetches, and continually updates, a subtree."""
        if self._require_client:
            scope.requires(self.client.scope)

        async with anyio.create_task_group() as tg:
            self._tg = tg

            async def monitor(*, task_status):
                pl = PathLongener(())
                await self.run_starting()
                async with self.client._stream(
                    "watch",
                    nchain=3,
                    path=self._path,
                    fetch=True,
                ) as w:
                    async for r in w:
                        if "path" not in r:
                            if r.get("state", "") == "uptodate":
                                await self.running()
                            task_status.started()
                            continue
                        pl(r)
                        val = r.get("value", NotGiven)
                        entry = self.follow(r.path, create=None, empty_ok=True)
                        if entry is not None:
                            # Test for consistency
                            try:
                                if entry.chain == r.chain:
                                    # entry.update() has set this
                                    await entry.seen_value()
                                elif _node_gt(entry.chain, r.chain):
                                    # stale data
                                    pass
                                elif not _node_gt(r.chain, entry.chain):
                                    entry.mark_inconsistent(r)
                            except AttributeError:
                                pass

                            # update entry
                            entry.chain = None if val is NotGiven else r.get("chain", None)
                            await entry.set_value(val)

                            if val is NotGiven and not entry:
                                # the entry has no value and no children,
                                # so we delete it (and possibly its
                                # parents) from our tree.
                                n = list(entry.subpath)
                                while n:
                                    # no-op except for class-specific side effects
                                    # like setting an event
                                    await entry.set_value(NotGiven)

                                    entry = entry.parent
                                    del entry[n.pop()]
                                    if entry:
                                        break

                        if not self._need_wait or "chain" not in r:
                            continue
                        c = r.chain
                        while c is not None:
                            if self._seen.get(c.node, 0) < c.tick:
                                self._seen[c.node] = c.tick
                            try:
                                w = self._waiters[c.node]
                            except KeyError:
                                pass
                            else:
                                while w and w[0][0] <= c.tick:
                                    heapq.heappop(w)[1].set()
                            c = c.get("prev", None)

            await tg.start(monitor)
            try:
                yield self
            finally:
                with anyio.fail_after(2, shield=True):
                    tg.cancel_scope.cancel()
            pass  # end of 'run', closing taskgroup

    async def cancel(self):
        """Stop the monitor"""
        await self._tg.cancel_scope.cancel()

    async def wait_loaded(self):
        """Wait for the tree to be loaded completely."""
        await self._loaded.wait()

    async def wait_chain(self, chain):
        """Wait for a tree update containing this tick."""
        try:
            if chain.tick <= self._seen[chain.node]:
                return
        except KeyError:
            pass
        w = self._waiters.setdefault(chain.node, [])
        e = anyio.Event()
        heapq.heappush(w, (chain.tick, e))
        await e.wait()

    async def spawn(self, p, *a, **kw):
        p = partial(p, *a, **kw)
        self._tg.start_soon(p)


class ClientRoot(MirrorRoot):
    """
    This class represents the root of a subsystem's storage.

    To use this class, create a subclass that, at minimum, overrides
    ``CFG`` and ``child_type``. ``CFG`` must be a dict with at least a
    ``prefix`` tuple. You instantiate the entry using :meth:`as_handler`.

    """

    def __init__(self, *a, **kw):
        if self.CFG is None:
            raise TypeError(f"You need to override .CFG in {type(self).__name__}")
        super().__init__(*a, **kw)



# from asyncclick.testing import CliRunner
from __future__ import annotations
import logging
import socket

import attr
from asyncscope import scope
from moat.src.test import run  # pylint:disable=import-error,no-name-in-module
from moat.util import (  # pylint:disable=no-name-in-module
    CFG,
    combine_dict,
    ensure_cfg,
)

from moat.kv.client import _scoped_client

logger = logging.getLogger(__name__)
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

ensure_cfg("moat.kv")


@attr.s
class S:
    tg = attr.ib()
    client_ctx = attr.ib()
    s = attr.ib(factory=list)  # servers
    c = attr.ib(factory=list)  # clients
    _seq = 1

    async def ready(self, i=None):
        if i is not None:
            await self.s[i].is_ready
            return self.s[i]
        for s in self.s:
            if s is not None:
                await s.is_ready
        return self.s

    def __iter__(self):
        return iter(self.s)

    @asynccontextmanager
    async def client(self, i: int = 0, **kv):
        """Get a (new) client for the i'th server."""
        await self.s[i].is_serving
        self._seq += 1
        for host, port, *_ in self.s[i].ports:
            if host != "::" and host[0] == ":":
                continue
            try:
                cfg = combine_dict(
                    dict(conn=dict(host=host, port=port, ssl=self.client_ctx, **kv)),
                    CFG["kv"],
                )

                async def scc(s, **cfg):
                    scope.requires(s._scope)
                    return await _scoped_client(scope.name, **cfg)

                async with scope.using_scope():
                    c = await scope.service(
                        f"moat.kv.client.{i}.{self._seq}",
                        scc,
                        self.s[i],
                        **cfg,
                    )
                    yield c
                return
            except socket.gaierror:
                pass
        raise RuntimeError("Duh? no connection")

    async def run(self, *args, do_stdout=True):
        h = p = None
        for s in self.s:
            for h, p, *_ in s.ports:
                if h[0] != ":":
                    break
            else:
                continue
            break
        if len(args) == 1:
            args = args[0]
            if isinstance(args, str):
                args = args.split(" ")
        async with scope.using_scope():
            return await run("-VV", "kv", "-h", h, "-p", p, *args, do_stdout=do_stdout)



#!/usr/bin/python3

# Batch-convert data. In this case I had some entries which were stored as
# a list, but using Path made much more sense (esp when you need to
# view/edit the yaml export).

import anyio
from moat.kv.client import open_client
from moat.util import P, yload, Path
import asyncclick as click


def conv(m, s: str) -> bool:
    try:
        d = m.value[s]
    except KeyError:
        return 0
    if isinstance(d, Path):
        return 0
    if not isinstance(d, Sequence):
        return 0
    d = Path.build(d)
    m.value[s] = d
    return 1


@click.command()
@click.argument("path", type=P)
@click.argument("keys", type=str, nargs=-1)
async def main(path, keys):
    if not keys:
        keys = "src dest dst state".split()
    with open("/etc/moat.kv.cfg") as cff:
        cfg = yload(cff)
    async with open_client(**cfg) as client:
        async for m in client.get_tree(path, nchain=2):
            n = 0
            for k in keys:
                n += conv(m, k)
            if n:
                await client.set(ORIG + m.path, value=m.value, chain=m.chain)


if __name__ == "__main__":
    main()



from __future__ import annotations
import copy
import logging
import os
import time
from contextlib import AsyncExitStack, asynccontextmanager
from functools import partial

import anyio
from unittest import mock
import trio
from asyncscope import main_scope, scope
from moat.mqtt.broker import create_broker
from moat.util import NotGiven, attrdict, combine_dict

from moat.kv.mock import S
from moat.kv.server import Server

from . import CFG

logger = logging.getLogger(__name__)

otm = time.time

PORT = 40000 + (os.getpid() + 10) % 10000

broker_cfg = {
    "listeners": {"default": {"type": "tcp", "bind": f"127.0.0.1:{PORT}"}},
    "timeout-disconnect-delay": 2,
    "auth": {"allow-anonymous": True, "password-file": None},
}

URI = f"mqtt://127.0.0.1:{PORT}/"


@asynccontextmanager
async def stdtest(n=1, run=True, ssl=False, tocks=20, **kw):
    C_OUT = CFG.get("_stdout", NotGiven)
    if C_OUT is not NotGiven:
        del CFG["_stdout"]
    TESTCFG = copy.deepcopy(CFG["kv"])
    TESTCFG.server.port = None
    TESTCFG.root = "test"
    if C_OUT is not NotGiven:
        CFG["_stdout"] = C_OUT
        TESTCFG["_stdout"] = C_OUT

    if ssl:
        import ssl

        import trustme

        ca = trustme.CA()
        cert = ca.issue_server_cert("127.0.0.1")
        server_ctx = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        client_ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
        ca.configure_trust(client_ctx)
        cert.configure_cert(server_ctx)
    else:
        server_ctx = client_ctx = False

    clock = trio.lowlevel.current_clock()
    try:
        clock.autojump_threshold = 0.02  # networking
    except Exception:
        pass  # test doesn't have autojump_clock fixture

    async def mock_get_host_port(st, host):
        i = int(host[host.rindex("_") + 1 :])
        s = st.s[i]
        await s.is_serving
        for host, port, *_ in s.ports:
            if host == "::" or host[0] != ":":
                return host, port

    def tm():
        try:
            return trio.current_time()
        except RuntimeError:
            return otm()

    async def mock_set_tock(self, old):
        assert self._tock < tocks, "Test didn't terminate. Limit:" + str(tocks)
        await old()

    done = False
    async with main_scope("moat.kv.test.mqtt") as scp:
        tg = scp._tg
        st = S(tg, client_ctx)
        async with AsyncExitStack() as ex:
            st.ex = ex  # pylint: disable=attribute-defined-outside-init
            ex.enter_context(mock.patch("time.time", new=tm))
            ex.enter_context(mock.patch("time.monotonic", new=tm))
            logging._startTime = tm()

            async def run_broker(cfg):
                async with create_broker(config=cfg) as srv:
                    # NB: some services use "async with await …"
                    scope.register(srv)
                    await scope.no_more_dependents()

            async def with_broker(s, *a, **k):
                await scope.service("moat.mqtt.broker", run_broker, broker_cfg)
                s._scope = scope.get()
                return await s._scoped_serve(*a, **k)

            args_def = kw.get("args", attrdict())
            for i in range(n):
                name = "test_" + str(i)
                args = kw.get(name, args_def)
                args = combine_dict(
                    args,
                    args_def,
                    {
                        "cfg": {
                            "conn": {"ssl": client_ctx},
                            "server": {
                                "bind_default": {
                                    "host": "127.0.0.1",
                                    "port": i + PORT + 1,
                                    "ssl": server_ctx,
                                },
                                "backend": "mqtt",
                                "mqtt": {"uri": URI},
                            },
                        },
                    },
                    {"cfg": TESTCFG},
                )
                args_def.pop("init", None)
                s = Server(name, **args)
                ex.enter_context(
                    mock.patch.object(s, "_set_tock", new=partial(mock_set_tock, s, s._set_tock)),
                )
                ex.enter_context(
                    mock.patch.object(s, "_get_host_port", new=partial(mock_get_host_port, st)),
                )
                st.s.append(s)

            evts = []
            for i in range(n):
                if kw.get(f"run_{i}", run):
                    evt = anyio.Event()
                    await scp.spawn_service(with_broker, st.s[i], ready_evt=evt)
                    evts.append(evt)
                else:
                    setattr(st, f"run_{i}", partial(scp.spawn_service, with_broker, st.s[i]))

            for e in evts:
                await e.wait()
            try:
                done = True
                yield st
            finally:
                with anyio.fail_after(2, shield=True):
                    logger.info("Runtime: %s", clock.current_time())
                    tg.cancel_scope.cancel()
    if not done:
        yield None



# command line interface
from __future__ import annotations

import datetime
import sys
from collections.abc import Mapping

import asyncclick as click
from moat.util import (
    MsgWriter,
    P,
    Path,
    load_subgroup,
    yload,
    yprint,
)

from moat.util.mqtt import unpacker
from moat.util.msgpack import StdMsgpack


@load_subgroup(short_help="Local data mangling", sub_pre="dump")
async def cli():
    """
    Low-level tools that don't depend on a running MoaT-KV server.
    """
    pass


@cli.command()
@click.argument("node", nargs=1)
@click.argument("file", type=click.Path(), nargs=1)
async def init(node, file):
    """Write an initial preload file.

    Usage: moat kv dump init <node> <outfile>

    Writes an initial MoaT-KV file that behaves as if it was generated by <node>.

    Using this command, followed by "moat kv server -l <outfile> <node>", is
    equivalent to running "moat kv server -i 'Initial data' <node>.
    """
    async with MsgWriter(path=file, codec="std-msgpack") as f:
        await f(
            dict(
                chain=dict(node=node, tick=1, prev=None),
                depth=0,
                path=[],
                tock=1,
                value="Initial data",
            ),
        )


@cli.command("msg")
@click.argument("path", nargs=1)
@click.pass_obj
async def msg_(obj, path):
    """
    Monitor the server-to-sever message stream.

    Use ':' for the main server's "update" stream.
    Use '+NAME' to monitor a different stream instead.
    Use '+' to monitor all streams.
    Otherwise use the given name as-is; Mosquitto wildcard rules apply.

    \b
    Common streams (prefix with '+'):
    * ping    all servers
    * update  data changes
    * del     nodes responsible for cleaning up deleted records
    """
    from moat.kv.backend import get_backend

    class _Unpack:
        def __init__(self):
            self._part_cache = dict()

    import moat.kv.server

    _Unpack._unpack_multiple = moat.kv.server.Server._unpack_multiple
    _unpacker = _Unpack()._unpack_multiple

    path = P(path)
    if len(path) == 0:
        path = P(obj.cfg.server["root"]) | "update"
    elif len(path) == 1 and path[0].startswith("+"):  # pylint: disable=no-member  # owch
        p = path[0][1:]
        path = P(obj.cfg.server["root"])
        path |= p or "#"
    be = obj.cfg.server.backend
    kw = obj.cfg.server[be]

    async with (
        get_backend(be)(**kw) as conn,
        conn.monitor(*path) as stream,
    ):
        async for msg in stream:
            v = vars(msg)
            if isinstance(v.get("payload"), (bytearray, bytes)):
                t = msg.topic
                v = unpacker(v["payload"])
                v = _unpacker(v)
                if v is None:
                    continue
                if not isinstance(v, Mapping):
                    v = {"_data": v}
                v["_topic"] = Path.build(t)
            else:
                v["_type"] = type(msg).__name__

            v["_timestamp"] = datetime.datetime.now().isoformat(
                sep=" ",
                timespec="milliseconds",
            )

            yprint(v, stream=obj.stdout)
            print("---", file=obj.stdout)


@cli.command("post")
@click.argument("path", nargs=1)
@click.pass_obj
async def post_(obj, path):
    """
    Send a msgpack-encoded message (or several) to a specific MQTT topic.

    Messages are read from YAML.
    Common streams:
    * ping: sync: all servers (default)
    * update: data changes
    * del: sync: nodes responsible for cleaning up deleted records
    """
    from moat.kv.backend import get_backend

    path = P(path)
    be = obj.cfg.server.backend
    kw = obj.cfg.server[be]

    async with get_backend(be)(codec=StdMsgpack(), **kw) as conn:
        for d in yload(sys.stdin, multi=True):
            topic = d.pop("_topic", path)
            await conn.send(*topic, payload=d)



# command line interface
from __future__ import annotations

import sys
import time
from functools import partial

import anyio
import asyncclick as click
from moat.util import P, Path, attr_args, attrdict, process_args, yprint

from moat.kv.code import CodeRoot
from moat.kv.data import add_dates, data_get
from moat.kv.runner import AllRunnerRoot, AnyRunnerRoot, SingleRunnerRoot


@click.group()  # pylint: disable=undefined-variable
@click.option("-n", "--node", help="node to run this code on. Empty: any one node, '-': all nodes")
@click.option("-g", "--group", help="group to run this code on. Empty: default")
@click.pass_context
async def cli(ctx, node, group):
    """Run code stored in MoaT-KV.

    \b
    The option '-n' is somewhat special:
    -n -     Jobs for all hosts
    -n XXX   Jobs for host XXX
    (no -n)  Jobs for any host

    The default group is 'default'.
    """
    obj = ctx.obj
    if group is None:
        group = "default"
    if group == "-":
        if node is not None:
            raise click.UsageError("'-g -' doesn't make sense with '-n'")
        if ctx.invoked_subcommand != "info":
            raise click.UsageError("'-g -' only makes sense with the 'info' command")
        obj.runner_root = SingleRunnerRoot
        subpath = (None,)
    elif not node:
        obj.runner_root = AnyRunnerRoot
        subpath = (group,)
    elif node == "-":
        obj.runner_root = AllRunnerRoot
        subpath = (group,)
    else:
        obj.runner_root = SingleRunnerRoot
        subpath = (node, group)

    cfg = obj.cfg["kv"]["runner"]
    obj.subpath = Path(cfg["sub"][obj.runner_root.SUB]) + subpath
    obj.path = cfg["prefix"] + obj.subpath
    obj.statepath = cfg["state"] + obj.subpath


@cli.group("at", short_help="path of the job to operate on", invoke_without_command=True)
@click.argument("path", nargs=1, type=P)
@click.pass_context
async def at_cli(ctx, path):
    """
    Add, list, modify, delete jobs at/under this path.
    """
    obj = ctx.obj
    obj.jobpath = path

    if ctx.invoked_subcommand is None:
        res = await obj.client.get(obj.path + path, nchain=obj.meta)
        yprint(
            res if obj.meta else res.value if "value" in res else None,
            stream=obj.stdout,
        )


@cli.command("info")
@click.pass_obj
async def info_(obj):
    """
    List available groups for the node in question.

    \b
    Options (between 'job' and 'info')
    (none)    list groups with jobs for any host
    -n -      list groups with jobs for all hosts
    -g -      list hosts that have specific jobs
    -n XXX    list groups with jobs for a specific host
    """
    path = obj.path[:-1]
    async for r in obj.client.get_tree(path=path, min_depth=1, max_depth=1, empty=True):
        print(r.path[-1], file=obj.stdout)


@at_cli.command("--help", hidden=True)
@click.pass_context
def help_(ctx):  # pylint:disable=unused-variable  # oh boy
    print(at_cli.get_help(ctx))


@at_cli.command("path")
@click.pass_obj
async def path__(obj):
    """
    Emit the full path leading to the specified runner object.

    Useful for copying or for state monitoring.

    NEVER directly write to the state object. It's controlled by the
    runner. You'll confuse it if you do that.

    Updating the control object will cancel any running code.
    """
    path = obj.jobpath
    res = dict(command=obj.path + path, state=obj.statepath + path)
    yprint(res, stream=obj.stdout)


@cli.command("run")
@click.option(
    "-n",
    "--nodes",
    type=int,
    default=0,
    help="Size of the group (not for single-node runners)",
)
@click.pass_obj
async def run(obj, nodes):
    """
    Run code that needs to run.

    This does not return.
    """
    from moat.util import as_service

    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")
    if nodes and obj.runner_root is SingleRunnerRoot:
        raise click.UsageError("A single-site runner doesn't have a size.")

    async with as_service(obj) as evt:
        c = obj.client
        cr = await CodeRoot.as_handler(c)
        await obj.runner_root.as_handler(
            c,
            subpath=obj.subpath,
            code=cr,
            **({"nodes": nodes} if nodes else {}),
        )
        evt.set()
        await anyio.sleep_forever()


async def _state_fix(obj, state, state_only, path, r):
    try:
        val = r.value
    except AttributeError:
        return
    if state:
        rs = await obj.client._request(
            action="get_value",
            path=state + r.path,
            iter=False,
            nchain=obj.meta,
        )
        if state_only:
            r.value = rs
        else:
            if obj.meta:
                val["state"] = rs
            elif "value" in rs:
                val["state"] = rs.value
        if "value" in rs:
            add_dates(rs.value)
    if not state_only:
        if path:
            r.path = path + r.path
        add_dates(val)

    return r


@at_cli.command("list")
@click.option("-s", "--state", is_flag=True, help="Add state data")
@click.option("-S", "--state-only", is_flag=True, help="Only output state data")
@click.option("-t", "--table", is_flag=True, help="one-line output")
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.pass_obj
async def list_(obj, state, state_only, table, as_dict):
    """List run entries."""
    if table and state:
        raise click.UsageError("'--table' and '--state' are mutually exclusive")

    path = obj.jobpath

    if state or state_only or table:
        state = obj.statepath + path

    if table:
        from moat.kv.errors import ErrorRoot

        err = await ErrorRoot.as_handler(obj.client)

        async for r in obj.client.get_tree(obj.path + path):
            p = path + r.path
            s = await obj.client.get(state + r.path)
            if "value" not in s:
                st = "-never-"
            elif s.value.started > s.value.stopped:
                st = s.value.node
            else:
                try:
                    e = await err.get_error_record("run", obj.path + p, create=False)
                except KeyError:
                    st = "-stopped-"
                else:
                    if e is None or e.resolved:
                        st = "-stopped-"
                    else:
                        st = " | ".join(
                            "{} {}".format(
                                Path.build(e.subpath)
                                if e._path[-2] == ee._path[-1]
                                else Path.build(ee.subpath),
                                getattr(ee, "message", None)
                                or getattr(ee, "comment", None)
                                or "-",
                            )
                            for ee in e
                        )
            print(p, r.value.code, st, file=obj.stdout)

    else:
        await data_get(
            obj,
            obj.path + path,
            as_dict=as_dict,
            item_mangle=partial(_state_fix, obj, state, state_only, None if as_dict else path),
        )


@at_cli.command("state")
@click.option("-r", "--result", is_flag=True, help="Just print the actual result.")
@click.pass_obj
async def state_(obj, result):
    """Get the status of a runner entry."""
    path = obj.jobpath

    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")
    if result and obj.meta:
        raise click.UsageError("You can't use '-v' and '-r' at the same time.")
    if not len(path):
        raise click.UsageError("You need a non-empty path.")
    path = obj.statepath + obj.jobpath

    res = await obj.client.get(path, nchain=obj.meta)
    if "value" not in res:
        if obj.debug:
            print("Not found (yet?)", file=sys.stderr)
        sys.exit(1)

    add_dates(res.value)
    if not obj.meta:
        res = res.value
    yprint(res, stream=obj.stdout)


@at_cli.command()
@click.option("-s", "--state", is_flag=True, help="Add state data")
@click.pass_obj
async def get(obj, state):
    """Read a runner entry"""
    path = obj.jobpath
    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")
    if not path:
        raise click.UsageError("You need a non-empty path.")

    res = await obj.client._request(
        action="get_value",
        path=obj.path + path,
        iter=False,
        nchain=obj.meta,
    )
    if "value" not in res:
        print("Not found.", file=sys.stderr)
        return
    res.path = path
    if state:
        state = obj.statepath
    await _state_fix(obj, state, False, path, res)
    if not obj.meta:
        res = res.value

    yprint(res, stream=obj.stdout)


@at_cli.command()
@click.option("-f", "--force", is_flag=True, help="Force deletion even if messy")
@click.pass_obj
async def delete(obj, force):
    """Remove a runner entry"""
    path = obj.jobpath

    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")
    if not path:
        raise click.UsageError("You need a non-empty path.")

    res = await obj.client.get(obj.path + path, nchain=3)
    if "value" not in res:
        res.info = "Does not exist."
    else:
        val = res.value
        if "target" not in val:
            val.target = None
        if val.target is not None:
            val.target = None
            res = await obj.client.set(obj.path + path, value=val, nchain=3, chain=res.chain)
            if not force:
                res.info = "'target' was set: cleared but not deleted."
        if force or val.target is None:
            sres = await obj.client.get(obj.statepath + path, nchain=3)
            if not force and "value" in sres and sres.value.stopped < sres.value.started:
                res.info = "Still running, not deleted."
            else:
                sres = await obj.client.delete(obj.statepath + path, chain=sres.chain)
                res = await obj.client.delete(obj.path + path, chain=res.chain)
                if "value" in res and res.value.stopped < res.value.started:
                    res.info = "Deleted (unclean!)."
                else:
                    res.info = "Deleted."

    if obj.meta:
        yprint(res, stream=obj.stdout)
    elif obj.debug:
        print(res.info)


@at_cli.command("set")
@click.option("-c", "--code", help="Path to the code that should run.")
@click.option("-C", "--copy", help="Use this entry as a template.")
@click.option("-t", "--time", "tm", help="time the code should next run at. '-':not")
@click.option("-r", "--repeat", type=int, help="Seconds the code should re-run after")
@click.option("-k", "--ok", type=float, help="Code is OK if it ran this many seconds")
@click.option("-b", "--backoff", type=float, help="Back-off factor. Default: 1.4")
@click.option("-d", "--delay", type=int, help="Seconds the code should retry after (w/ backoff)")
@click.option("-i", "--info", help="Short human-readable information")
@attr_args
@click.pass_obj
async def set_(obj, code, tm, info, ok, repeat, delay, backoff, copy, **kw):
    """Add or modify a runner.

    Code typically requires some input parameters.

    You should use '-v NAME VALUE' for string values, '-p NAME VALUE' for
    paths, and '-e NAME VALUE' for other data. '-e NAME -' deletes an item.
    """
    path = obj.jobpath

    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")

    if code is not None:
        code = P(code)
    if copy:
        copy = P(copy)
    path = obj.path + P(path)

    res = await obj.client._request(action="get_value", path=copy or path, iter=False, nchain=3)
    if "value" not in res:
        if copy:
            raise click.UsageError("--copy: use the complete path to an existing entry")
        elif code is None:
            raise click.UsageError("New entry, need code")
        res = {}
        chain = None
    else:
        chain = None if copy else res["chain"]
        res = res["value"]
        if copy and "code" not in res:
            raise click.UsageError("'--copy' needs a runner entry")

    vl = attrdict(**res.setdefault("data", {}))
    vl = process_args(vl, **kw)
    res["data"] = vl

    if code is not None:
        res["code"] = code
    if ok is not None:
        res["ok_after"] = ok
    if info is not None:
        res["info"] = info
    if backoff is not None:
        res["backoff"] = backoff
    if delay is not None:
        res["delay"] = delay
    if repeat is not None:
        res["repeat"] = repeat
    if tm is not None:
        if tm == "-":
            res["target"] = None
        else:
            res["target"] = time.time() + float(tm)

    res = await obj.client.set(path, value=res, nchain=3, chain=chain)
    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.command(short_help="Show runners' keepalive messages")
@click.pass_obj
async def monitor(obj):
    """
    Runners periodically send a keepalive message. Show them.
    """

    # TODO this does not watch changes in MoaT-KV.
    # It also should watch individual jobs' state changes.
    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")

    async with obj.client.msg_monitor("run") as cl:
        async for msg in cl:
            yprint(msg, stream=obj.stdout)
            print("---", file=obj.stdout)



# command line interface
from __future__ import annotations

import datetime
import time

import asyncclick as click
from moat.util import MsgReader, NotGiven, P, PathLongener, attr_args, yprint
from moat.util.times import ts2iso

from moat.kv.client import StreamedRequest
from moat.kv.data import add_dates, data_get, node_attr


@click.group(short_help="Manage data.", invoke_without_command=True)  # pylint: disable=undefined-variable
@click.argument("path", type=P, nargs=1)
@click.pass_context
async def cli(ctx, path):
    """
    This subcommand accesses the actual user data stored in your MoaT-KV tree.
    """
    if ctx.invoked_subcommand is None:
        await data_get(ctx.obj, path, recursive=False)
    else:
        ctx.obj.path = path


@cli.command()
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m",
    "--maxdepth",
    type=int,
    default=None,
    help="Limit recursion depth. Default: whole tree",
)
@click.option(
    "-M",
    "--mindepth",
    type=int,
    default=None,
    help="Starting depth. Default: whole tree",
)
@click.option("-r", "--recursive", is_flag=True, help="Read a complete subtree")
@click.option("-e", "--empty", is_flag=True, help="Include empty nodes")
@click.option("-R", "--raw", is_flag=True, help="Print string values without quotes etc.")
@click.option("-D", "--add-date", is_flag=True, help="Add *_date entries")
@click.pass_obj
async def get(obj, **k):
    """
    Read a MoaT-KV value.

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """

    await data_get(obj, obj.path, **k)


@cli.command("list")
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m",
    "--maxdepth",
    type=int,
    default=1,
    help="Limit recursion depth. Default: 1 (single layer).",
)
@click.option(
    "-M",
    "--mindepth",
    type=int,
    default=1,
    help="Starting depth. Default: 1 (single layer).",
)
@click.pass_obj
async def list_(obj, **k):
    """
    List MoaT-KV values.

    This is like "get" but with "--mindepth=1 --maxdepth=1 --recursive --empty"

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """

    k["recursive"] = True
    k["raw"] = True
    k["empty"] = True
    await data_get(obj, obj.path, **k)


@cli.command("set", short_help="Add or update an entry")
@attr_args
@click.option("-l", "--last", nargs=2, help="Previous change entry (node serial)")
@click.option("-n", "--new", is_flag=True, help="This is a new entry.")
@click.pass_obj
async def set_(obj, last, new, **kw):
    """
    Store a value at some MoaT-KV position.

    If you update a value you can use "--last" to ensure that no other
    change arrived between reading and writing the entry. (This is not
    foolproof but reduces the window to a fraction of a second.)

    When adding a new entry use "--new" to ensure that you don't
    accidentally overwrite something.

    MoaT-KV entries typically are mappings. Use a colon as the path if you
    want to replace the top level.
    """
    if new:
        if last:
            raise click.UsageError("'new' and 'last' are mutually exclusive")
        kw["chain"] = None
    else:
        if last:
            kw["chain"] = {"node": last[0], "tick": int(last[1])}

    res = await node_attr(obj, obj.path, **kw)

    if obj.meta:
        yprint(res, stream=obj.stdout)


class nstr:
    def __new__(cls, val):
        if val is NotGiven:
            return val
        return str(val)


@cli.command(short_help="Delete an entry / subtree")
@click.option(
    "-p",
    "--prev",
    type=nstr,
    default=NotGiven,
    help="Previous value. Deprecated; use 'last'",
)
@click.option("-l", "--last", nargs=2, help="Previous change entry (node serial)")
@click.option("-r", "--recursive", is_flag=True, help="Delete a complete subtree")
@click.option("--internal", is_flag=True, help="Affect the internal tree. DANGER.")
@click.option("-e", "--eval", "eval_", is_flag=True, help="The previous value shall be evaluated.")
@click.pass_obj
async def delete(obj, prev, last, recursive, eval_, internal):
    """
    Delete an entry, or a whole subtree.

    You really should use "--last" (preferred) or "--prev" (if you must) to
    ensure that no other change arrived (but note that this doesn't work
    when deleting a subtree).

    Non-recursively deleting an entry with children works and does *not*
    affect the child entries.

    The root entry cannot be deleted.
    """
    args = {}
    if eval_ and prev is NotGiven:
        raise click.UsageError("You need to add a value that can be evaluated")
    if recursive:
        if prev is not NotGiven or last:
            raise click.UsageError("You can't use a prev value when deleting recursively.")
        if internal:
            raise click.UsageError("'internal' and 'recursive' are mutually exclusive")
    else:
        if prev is not NotGiven:
            if eval_:
                prev = eval(prev)  # pylint: disable=eval-used
            args["prev"] = prev
        if last:
            args["chain"] = {"node": last[0], "tick": int(last[1])}

    res = await obj.client.delete(path=obj.path, nchain=obj.meta, recursive=recursive, **args)
    if isinstance(res, StreamedRequest):
        pl = PathLongener(obj.path)
        async for r in res:
            pl(r)
            if obj.meta:
                yprint(r, stream=obj.stdout)
    else:
        if obj.meta:
            yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-s", "--state", is_flag=True, help="Also get the current state.")
@click.option("-o", "--only", is_flag=True, help="Value only, nothing fancy.")
@click.option("-p", "--path-only", is_flag=True, help="Value only, nothing fancy.")
@click.option("-D", "--add-date", is_flag=True, help="Add *_date entries")
@click.option("-i", "--ignore", multiple=True, type=P, help="Skip this (sub)tree")
@click.pass_obj
async def monitor(obj, state, only, path_only, add_date, ignore):
    """Monitor a MoaT-KV subtree"""

    flushing = not state
    seen = False

    async with obj.client.watch(
        obj.path,
        nchain=obj.meta,
        fetch=state,
        max_depth=0 if only else -1,
        long_path=False,
    ) as res:
        pl = PathLongener(() if path_only else obj.path)
        async for r in res:
            if add_date and "value" in r:
                add_dates(r.value)
            if any(p == r.path[: len(p)] for p in ignore):
                continue
            pl(r)
            if r.get("state", "") == "uptodate":
                if only and not seen:
                    # value doesn't exist
                    return
                flushing = True
                if only or path_only:
                    continue
            else:
                del r["seq"]
                if only or path_only:
                    try:
                        if path_only:
                            print(f"{r.path} {r.value}", file=obj.stdout)
                        else:
                            print(r.value, file=obj.stdout)
                        continue
                    except AttributeError:
                        # value has been deleted
                        continue
            if flushing:
                r["time"] = ts = time.time()
                r["_time"] = ts2iso(ts)
            yprint(r, stream=obj.stdout)
            print("---", file=obj.stdout)
            if flushing:
                obj.stdout.flush()
            seen = True


@cli.command()
@click.option("-i", "--infile", type=click.Path(), help="File to read (msgpack).")
@click.pass_obj
async def update(obj, infile):
    """Send a list of updates to a MoaT-KV subtree"""
    async with MsgReader(path=infile, codec="std-msgpack") as reader:
        async for msg in reader:
            if not hasattr(msg, "path"):
                continue
            await obj.client.set(obj.path + msg.path, value=msg.value)



#
"""
Null auth method.

Does not limit anything, allows everything.
"""

from __future__ import annotations

from . import (
    BaseClientAuth,
    BaseClientAuthMaker,
    BaseServerAuthMaker,
    RootServerUser,
    null_client_login,
    null_server_login,
)


def load(typ: str, *, make: bool = False, server: bool):
    if typ == "client":
        if server:
            return null_server_login
        else:
            return null_client_login
    if typ != "user":
        raise NotImplementedError("This module only handles users")
    if server:
        if make:
            return ServerUserMaker
        else:
            return ServerUser
    else:
        if make:
            return ClientUserMaker
        else:
            return ClientUser


class ServerUserMaker(BaseServerAuthMaker):
    schema = {"type": "object", "additionalProperties": False}


class ServerUser(RootServerUser):
    schema = {"type": "object", "additionalProperties": False}


class ClientUserMaker(BaseClientAuthMaker):
    gen_schema = {"type": "object", "additionalProperties": False}
    mod_schema = {"type": "object", "additionalProperties": False}

    @property
    def ident(self):
        return "*"


class ClientUser(BaseClientAuth):
    schema = {"type": "object", "additionalProperties": False}



# command line interface
from __future__ import annotations

import json

import asyncclick as click
from moat.util import NotGiven, P, Path, PathLongener, yload, yprint
import contextlib


@click.group()  # pylint: disable=undefined-variable
async def cli():
    """Manage types and type matches. Usage: … type …"""
    pass


@cli.command()
@click.option("-s", "--script", type=click.File(mode="w", lazy=True), help="Save the script here")
@click.option("-S", "--schema", type=click.File(mode="w", lazy=True), help="Save the schema here")
@click.option("-y", "--yaml", "yaml_", is_flag=True, help="Write schema as YAML. Default: JSON.")
@click.argument("path", type=P, nargs=1)
@click.pass_obj
async def get(obj, path, script, schema, yaml_):
    """Read type checker information"""
    if not len(path):
        raise click.UsageError("You need a non-empty path.")
    res = await obj.client._request(
        action="get_internal",
        path=Path("type") + path,
        iter=False,
        nchain=obj.meta,
    )
    try:
        r = res.value
    except AttributeError:
        raise click.UsageError(f"No data at {Path('type') + path}") from None

    if not obj.meta:
        res = res.value
    if script:
        script.write(r.pop("code"))
    if schema:
        if yaml_:
            yprint(r.pop("schema"), stream=schema)
        else:
            json.dump(r.pop("schema"), schema)
    yprint(res, stream=obj.stdout)


@cli.command("set")
@click.option("-g", "--good", multiple=True, help="Example for passing values")
@click.option("-b", "--bad", multiple=True, help="Example for failing values")
@click.option("-d", "--data", type=click.File(mode="r"), help="Load metadata from this YAML file.")
@click.option("-s", "--script", type=click.File(mode="r"), help="File with the checking script")
@click.option("-S", "--schema", type=click.File(mode="r"), help="File with the JSON schema")
@click.option("-y", "--yaml", "yaml_", is_flag=True, help="load the schema as YAML. Default: JSON")
@click.argument("path", type=P, nargs=1)
@click.pass_obj
async def set_(obj, path, good, bad, script, schema, yaml_, data):
    """Write type checker information."""
    if not len(path):
        raise click.UsageError("You need a non-empty path.")

    if data:
        msg = yload(data)
    else:
        msg = {}
    chain = NotGiven
    if "value" in msg:
        chain = msg.get("chain", NotGiven)
        msg = msg["value"]

    msg.setdefault("good", [])
    msg.setdefault("bad", [])
    for x in good:
        msg["good"].append(eval(x))  # pylint: disable=eval-used
    for x in bad:
        msg["bad"].append(eval(x))  # pylint: disable=eval-used

    if "code" in msg:
        if script:
            raise click.UsageError("Duplicate script")
    elif script:
        msg["code"] = script.read()

    if "schema" in msg:
        raise click.UsageError("Missing schema")
    elif schema:
        if yaml_:
            msg["schema"] = yload(schema)
        else:
            msg["schema"] = json.load(schema)

    if "schema" not in msg and "code" not in msg:
        raise click.UsageError("I need a schema, Python code, or both.")

    if len(msg["good"]) < 2:
        raise click.UsageError("Missing known-good test values (at least two)")
    if not msg["bad"]:
        raise click.UsageError("Missing known-bad test values")

    res = await obj.client._request(
        action="set_internal",
        value=msg,
        path=Path("type") + path,
        iter=False,
        nchain=obj.meta,
        **({} if chain is NotGiven else {"chain": chain}),
    )
    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-R", "--raw", is_flag=True, help="Print just the path.")
@click.option("-t", "--type", "type_", help="Type path to link to.")
@click.option("-d", "--delete", help="Use to delete this mapping.")
@click.argument("path", type=P, nargs=1)
@click.pass_obj
async def match(obj, path, type_, delete, raw):  # pylint: disable=redefined-builtin
    """Match a type to a path (read, if no type given; list if empty path)"""
    if not len(path):
        if raw or type_ or delete:
            raise click.UsageError("No options allowed when dumping the match tree..")
        y = {}
        pl = PathLongener()
        async for r in await obj.client._request(
            "get_tree_internal",
            path=Path("match") + path,
            iter=True,
            nchain=0,
        ):
            pl(r)
            path = r["path"]
            yy = y
            for p in path:
                yy = yy.setdefault(p, {})
            with contextlib.suppress(KeyError):
                yy["_"] = r["value"]
        yprint(y, stream=obj.stdout)
        return

    if type_ and delete:
        raise click.UsageError("You can't both set and delete a path.")
    if raw and (type_ or delete):
        raise click.UsageError("You can only print the raw path when reading a match.")

    if delete:
        res = await obj.client._request(action="delete_internal", path=Path("type") + path)
        if obj.meta:
            yprint(res, stream=obj.stdout)
        return

    msg = {}
    if type_:
        msg["type"] = P(type_)
        act = "set_internal"
    elif delete:
        act = "delete_internal"
    else:
        act = "get_internal"
    res = await obj.client._request(
        action=act,
        value=msg,
        path=Path("match") + path,
        iter=False,
        nchain=obj.meta,
    )
    if obj.meta:
        yprint(res, stream=obj.stdout)
    elif type_ or delete:
        pass
    else:
        print(" ".join(str(x) for x in res.type), file=obj.stdout)


@cli.command()
@click.argument("path", type=P, nargs=1)
@click.pass_obj
async def list(obj, path):  # pylint: disable=redefined-builtin
    """Dump type data"""

    y = {}
    pl = PathLongener()
    async for r in await obj.client._request(
        "get_tree_internal",
        path=Path("type") + path,
        iter=True,
        nchain=0,
    ):
        pl(r)
        path = r["path"]
        yy = y
        for p in path:
            yy = yy.setdefault(p, {})
        with contextlib.suppress(KeyError):
            yy["_"] = r["value"]
    yprint(y, stream=obj.stdout)



from __future__ import annotations
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager

import anyio

__all__ = ["get_backend", "Backend"]


class Backend(metaclass=ABCMeta):
    def __init__(self, tg):
        self._tg = tg
        self._njobs = 0
        self._ended = None

    @abstractmethod
    @asynccontextmanager
    async def connect(self, *a, **k):
        """
        This async context manager returns a connection.
        """

    async def aclose(self):
        """
        Force-close the connection.
        """
        self._tg.cancel_scope.cancel()
        if self._njobs > 0:
            with anyio.move_on_after(2):
                await self._ended.wait()

    async def spawn(self, p, *a, **kw):
        async def _run(p, a, kw, *, task_status):
            if self._ended is None:
                self._ended = anyio.Event()
            self._njobs += 1
            task_status.started()
            try:
                return await p(*a, **kw)
            finally:
                self._njobs -= 1
                if not self._njobs:
                    self._ended.set()
                    self._ended = None

        return await self._tg.start(_run, p, a, kw)

    @abstractmethod
    @asynccontextmanager
    async def monitor(self, *topic):
        """
        Return an async iterator that listens to this topic.
        """

    @abstractmethod
    async def send(self, *topic, payload):
        """
        Send this payload to this topic.
        """


def get_backend(name):
    from importlib import import_module

    if "." not in name:
        name = "moat.kv.backend." + name
    return import_module(name).connect
